(* struct Check *)

open Output
open Pie
open Exp
open Surr
open Norm

(* Options *)

let indent = Output.indent
let verbWfT = true (* give trace for constructor checking *)

(* Type Checking with EVars *)

exception Fail of string


(** Types for CPS-style type-checking *)

type result = env * (env * exp)
(** A [result] is a pair [(thR, (thL, e))] of a type environment 
    and a Church term closure. *)

(*
type cont  = result -> result
 A continuation is a [result] transformer. *)


type level = 
    LTerm (* check term against type *)
  | LType (* check type operator against kind *)
  | LKind (* check wellformedness of kind *)


let assign (i, c, ((theta, gamma), rho)) =
    match !c with
	Name (x, gX, kappa) ->
	  (* print_string "ASSIGN"; *)
	  (* chkConstr (((theta, gX), rho), kappa); *)
	  let rho' = promote (gamma, rho, gX)
	  in
	    log ("set_chk", fun () ->
		   "set "^(String.make i ' ')^x^" = "^Clos.toString ((theta, gX), rho')^"\n");
	    c := Value (rho', gX, kappa)


let rec chkVerb' (lev, verb, cL0, cR0) : result =
(** [chkVerb' (lev, verb, cL0, cR0)]
   Input: [lev]     = [LTerm | LType | LKind]
          [verb]    : display progress of type checking?
          [cL0, cR0]: left and right hand side of checking relation
*)
           

(* Normalization strategy
wnf : - at the beginning
      - after instantiation of EVar (e.g. unification)
hnf : - before chk
*)


(* unused
(* tryEVar (expand, (((th0, g), tau0) as c0)) = (th1, tau1, ex, x) option
**
** Input:  c0     : closure in weak normal form
**         expand : expand strict definitions ?
** Output: th1    : (poss.) extended environment th0
**         tau1   : (poss.) tau0 with leading Forall stripped
**         ex     : (poss.) Church term
**         x      : (poss.) name of new EVar
*)
let rec tryEVar (expand, (((theta0, gamma), tau0) as c0)) =
  let rec iter' (theta, tau) =
    match tau with
	Var j -> 
	  (if j < 0 then 
	     match env_lookup (theta, j) with
		 Decl (x, kappa) -> None
	       | Def (x, kappa, sigma, strict) -> 
		   if expand || not strict then iter (theta, sigma)
		   else None
	       | Cell c -> match !c with
		     Name (x, gX, kappa) -> None
		   | Value (rho, gX, kappa) -> iter (theta, promote (gX, rho, gamma))
	   else None)
      | C(Forall, [[],kappa; [a],sigma]) -> 
	   Some (env_introEVarName (a, gamma, kappa, theta, sigma))
      | _ -> None
  and iter (theta, tau) =
    let tau' = if expand then hnf ((theta, gamma), tau) else tau
    in iter' (theta, tau')
  in
    iter (theta0, tau0) 
uncomment tryEVar *)

let notFunctionType (cl) =
  raise (Fail ("The type "^Clos.toString (cl)^" is not a function type.\n")) 
and illFormed (cl) =
  raise (Error ("Encountered illegal expression '"^Clos.toString (cl)^"'.\n"))
in

(* eVarToArr (ind, (th, g), X vec) = ((th[X=Y->Z,Y=?,Z=?], g), Y vec, Z vec) *)

let eVarToArr (ind, ((theta, gamma), tau0)) =

  let fail () = notFunctionType ((theta, gamma), tau0) in

  let rec mkArr' (n, y, z) =
    if (n = 0) then mkArr (y, z)
    else C(Lam, [["_"], mkArr' (n-1, mkApp (y, Var (n-1)), mkApp (z, Var (n-1)))])
  in
  let rec iter (n, t) =
    match t with
	Var i -> if i >= 0 then notFunctionType ((theta, gamma), tau0)
	else (match env_lookup (theta, i) with
		  Cell c ->
		    (match !c with
			 Name (x, gX, kappa) -> 
			   let (th, y) = env_newEVar (x, gamma, kappa, theta)
			   in let (th', z) = env_newEVar (x, gamma, kappa, th)
			   in assign (ind, c, ((th', gamma), mkArr' (n, y, z)));
			     (th', y, z) 
			     
		       | Value (sigma, gX, kappa) -> iter (n, promote (gamma, sigma, gX)))
	  	| _ -> fail ())
      | C(App, [[],rho; [],sigma]) ->
	  let (th, y, z) = iter (n+1, rho) 
	  in (th, mkApp (y, sigma), mkApp (z, sigma)) (* X -> Y for higher kinds *)
(*      | C(Arr, [[],rho; [],sigma]) -> (theta, rho, sigma) *)
      | _ -> fail ()

  in (* begin eVarToArr *)
    iter (0, tau0)

(* chk' (info, cL, cR) = (th, e)
**
** Input:  info   for printing
**         cL, cR in weak normal form (no redices)
** Output: th : updated enviroment thR
**         e  : Church term
*)

in
let rec chk' (((first, i) as info), (((thL, gL), tL) as cL), (((thR, gR), tR) as cR)) : result =

  (* keep a copy of thR for post-lifting EVars in subtype-checking *)
  let thRstop = thR in

  if verb (* lev = LTerm *) then 
    log ("enter_chk", fun () -> 
	   "CHK "^(Clos.toString cL)^" : "^(Clos.toString cR)^"\n") 
  else ();


  let assume (surr, a, kappa) k =
    chkKind (kappa);
    let newI = if first then i+indent else i
    in let surr' = Clos.pushTyVar (a, kappa, surr)
    in
      if verb then
	log ("ass_chk", fun () ->
	     "ass "^(String.make newI ' ')^
	     (Clos.toString (surr', Var 0))^" : "^
	     (Clos.toString (Clos.close (kappa)))^"\n")
      else ();
      k ((false, newI), surr')
  in

  let forallR (a, kappa, sigma) =
    assume ((thR, gR), a, kappa)
      (fun (info', sR') ->
	 let (theta', (thL', e)) = chk' (info', cL, (sR', sigma))
	 in
	   if first then 
	     log ("chk_chk", fun () -> "chk "^(String.make i ' ')^
		  (Clos.toString cL)^" : "^
		  (Clos.toString (wnfc ((theta', gR), tR)))^"\n")
	   else ();
	   (theta', (thL', mkTyAbs (a, kappa, e))))
	
  in (* begin chk' *)

    match (tL, tR) with

(*  eager introduction of universal vars will be unnecessary
    when using subtyping in unification *)

      (* welltyped terms: Lambda + *)
(*
    | (_, C(Forall, [[],kappa; [a],sigma])) -> forallR (a, kappa, sigma)
*)

      (* wellformed lambdas (L): wfType and wtTerm *)
    | (C(Lam, [[x],r]), _) ->

	(match (hnf cR) with        (* expose constructor *)
	     
	     C(Arr, [[],rho;[],sigma]) -> 

	       (* rho must be a type or a kind *)
	       if (lev = LTerm) then chkType' ((thR, gR), rho) else
	       if (lev = LType) then chkKind (rho) else
               illFormed (cL);

	       let newI = if first then i+indent else i
	       in let sL' = Clos.pushTmVar (x, (gR, rho), (thL, gL))
               in 
		 if verb (* lev = LTerm *) then
		   log ("ass_chk", fun () ->
			  "ass "^(String.make newI ' ')^
			  (Clos.toString (sL', Var 0))^" : "^
			  (Clos.toString ((thR, gR), rho))^"\n")
		 else ();
		 let (theta', (thL', e)) = 
		   chk' ((false, newI), (sL', r), ((thR, gR), sigma))
				     (* in let rho' = wnf ((theta', gR), rho) *)
		 in
		   if verb (* lev = LTerm *) && first then 
		     log ("chk_chk", fun () ->
			    "chk "^(String.make i ' ')^
			    (Clos.toString cL)^" : "^
			    (Clos.toString (wnfc ((theta', gR), tR)))^"\n")
		   else ();
		   (theta', (thL', mkTmAbs (x, rho, e)))
	
	   | C(Forall, [[],kappa; [a],sigma]) -> forallR (a, kappa, sigma)
		   
	   (* wellformed lambdas (L): wtTerm of Type X vec *)
	   | _ -> 
	       let (thR', y, z) = eVarToArr (i, cR)
	       in chk' (info, cL, wnfc ((thR', gR), mkArr (y, z)))
	)

      (* wellformed arrows (L): wfKind and wfType *)
    | (C(Arr, [[],s;[],t]), _) -> 
	(* e1 -> e2 : e;  e must be "*" or "kind" *)
	if lev = LTerm then illFormed (cL) else
	  (ignore (chk' (info, ((thL, gL), s), cR)); 
	   chk' (info, ((thL, gL), t), cR))

      (* wellformed types: Arr + Lambda + *)
    | (C (Forall, [[],kappa; [a],tau]), C(Type, [])) ->
	assume ((thL, gL), a, kappa)
	(fun (info', sL') ->
	   chk' (info', (sL', tau), cR))
    
    | (C (Forall, _), _) ->
	if lev = LType then 
	  raise (Fail ("The universal quantification "^
		       Clos.toString ((thL, gL), tL)^
		       " is expected to be of kind "^Keyword.star^
		       ", but here claimed to be of kind "^
		       Clos.toString ((thR, gR), tR)^".\n"))
	else illFormed (cL)

      (* wellformed kinds [cR = (Clos.empty, Exp.mkKind)]: Arr + *)
    | (C(Type, []), C(Kind, [])) -> (thR, (thL, mkType))


      (* else  chk ((gL, thL), x @ ts) : cR
         - first collect ts and type of x 
         - compute types of the ts and of x @ ts
         - unify computed type with cR
         - check types of ts *)
    | _ ->

      (* [doArgs : (bool * int) * bool * exp list * exp -> result  -> ((env * cxt) * env -> env * env) -> result]
	 CPS-style handling of args ts
         modTh : bool  ? has th been modified (to avoid infinite loops)
         ts    : args to x which still have to be processed
         rho   : remaining type of x after application of some args
         k     : continuation: what to do after all ts are done 
       *)
        let rec doArgs ((first, i) as info, modTh, ts, rho) (thR, ((thL, e) as ecl)) (k :(env * cxt) * env -> env * env)  =

	  let error () = 
	    (* if [rho] is not an arrow, instead of just raising 
	       an error, try to obtain more info and check 
	       types of args [ts] so far.  Then continue *)
	    if modTh then 
	      let (thR', thL') = (k ((thR, gR), thL))
	      in 
		doArgs (info, false, ts, rho) (thR', (thL', e)) 
		   (fun ((thR', gR'), thL') -> (thR', thL'))
	    else notFunctionType ((thR, gR), rho)

	  and contWithNewEVar (thR', tau, ex, x) =
	    let newI = if first then i+indent else i in
	    log ("new_chk", fun () -> "new "^(String.make newI ' ')^x^"\n");
	    doArgs ((false, newI), modTh, ts, tau) 
	      (thR', (thL, mkTyApp (e, ex))) k

	  in
	  match ts with
	      [] -> (* eagerly introduce existential variables, but
                       expand only non-strict definitions *)

(* comment tryEVar
		(match tryEVar (false, wnfc ((thR, gR), rho)) with
		     Some (thR', tau, ex, x) -> 
		       contWithNewEVar (thR', tau, ex, x)
(* org. version
		       let k' (thR, e) = 
			 let (thR, e) = k (thR, e)
			 in (thR, mkTyApp (e, x))
		       in doArgs (modTh, ts, tau) (thR', e) k' 
 *)
(* failed fix
		       let k' (thR, e) = k (thR, mkTyApp (e, x))
		       in doArgs (modTh, ts, tau) k' (thR', e)
*)
(* latest version: works [for open0 in exF.syf] 
		       doArgs (modTh, ts, tau) (thR', (thL, mkTyApp (e, x))) k
*)
		   | None ->
end uncomment tryEVar *)

                       if (lev = LTerm) then

			 (* check whether to defer unification problem 
			    because it could be too hard *)
			 if modTh && Uni.isFlexPlusArgs ((thR, gR), rho) then
			   error ()
			 else

		           (log ("uni_chk", fun () ->
				 "uni "^(String.make i ' ')^
				 (Clos.toString ((thR, gR), rho))^" <= "^
				 (Clos.toString (wnfc ((thR, gR), tR)))^"\n");
			    log ("church_chk", fun () ->
				 (String.make i ' ')^"      [Church-in: " ^
				 Church.toString ((thL, gL), (thR, gR), e)^"]\n");
			    let (gR', (thR', e')) =
			      Uni.subType (i, thRstop) ((thL, gL), (thR, gR), e) (rho, tR)
			    in
			      log ("chk_chk", fun () ->
				   "chk "^(String.make i ' ')^
			       	   (Clos.toString (cL))^" : "^
			       	   (Clos.toString (wnfc ((thR', gR), tR)))^
			       	   "\n");
			      (* invoke [k]: checking of arg types *)
			      let (thR2, thL2) = k ((thR', gR'), thL)
			      in  (thR2, (thL2, e')))
			   
		       else (* lev = LType: only check equality of types *)
			 (if verb && modTh (* at least one arg? *) then
			    log ("chk_chk", fun () -> "?ki "^(String.make i ' ')^
			       	   (Clos.toString (cL))^" : "^
			       	   (Clos.toString (wnfc ((thR,gR), tR)))^"\n")
(*		           log ("eq_chk", fun () ->
				"eq "^(String.make i ' ')^
				(Clos.toString ((thR, gR), rho))^" == "^
				(Clos.toString (wnfc ((thR, gR), tR)))^"\n")
*)			     
			 else ();
			 Eq.eq ((thR, gR), rho, tR, mkKind);
(*
			  if verb then
			      log ("chk_chk", fun () -> "chk "^(String.make i ' ')^
			       	   (Clos.toString (cL))^" : "^
			       	   (Clos.toString (wnfc ((thR,gR), tR)))^"\n")
			  else ();
*)
			  let (thR2, thL2) = k ((thR, gR), thL)
			  in  (thR2, (thL2, e)))


	    | t :: ts' -> 
		match (hnf ((thR, gR), rho)) with

		    C(Arr, [[],sigma; [],tau]) -> 

		      (* 1st step: insert placeholder in Church term *)
		      let (thL', ex, r) = env_newChVar ("e", gL, gR, sigma, thL) in
			if lev = LTerm then 
			  log ("church_uni", fun () -> 
				 let ChName (x, _, _, _) = !r in
				   "new "^(String.make i ' ')^x^"\n")
			else ();

			(* 2nd step: continue [doArgs] with ext. Ch.term *) 
			doArgs (info, true, ts', tau) (thR, (thL', mkTmApp (e, ex)))
			  (fun ((thR1, gR1), thL1) ->
			     
			     (* raise (Error ("entering cont")); *)

			     (* 3nd step: k *)
			     let (thR2, thL2) = k ((thR1, gR1), thL1) in

			       (* raise (Error ("did some cont")); *)
			       
			       (* 4th step: check arg. *)
			     let sigma' = promote (gR, sigma, gR1) in 
			     let (thR3, (thL3, ex')) =
			       chk' ((true, i), ((thL2, gL), t), ((thR2, gR1), sigma')) in
			       (* raise (Error ("did some check")); *)
				 (* 5th step: assign result Ch.t. to placeh. *)
			       if lev = LTerm then 
				 log ("church_uni", fun () -> 
					let ChName (x, _, _, _) = !r in
					  "set "^(String.make i ' ')^x^
					  " = "^(Church.toString ((thL3, gL), (thR3, gR1), ex'))^"\n") 
			       else ();

			       r := ChValue (ex', gL, gR1, sigma');
(*
			       if lev = LTerm then (let Name (x, _, _) = !r in
				  print_string ("set "^(String.make i ' ')^x^
						" = "^(Church.toString ((thL3, gL), (thR3, gR), ex'))^"\n")) 
			       else ();

			       r := Value (ex', gL, sigma);
*)
			       (thR3, thL3))
			     
(* original version
		      let k' (thR, e) =
			let (thR, e) = k (thR, e)
			in let (thR', e') = 
			  chk' ((false, i), ((thL, gL), t), ((thR, gR), sigma))
			in (thR', mkTmApp (e, e'))
		      in doArgs (true, ts', tau) k' (thR, e)
*)
		  | C(Forall, [[],kappa; [a],tau]) ->
		      contWithNewEVar (env_introEVarName (a,gR,kappa, thR, tau))
(*
		      let (thR', tau', ex, x) = 
			env_introEVarName (a,gR,kappa, thR, tau) in
			contWithNewEVar (thR', tau', ex, x)
*)		      
(* original version
		      let k' (thR, e) = 
			  let (thR, e) = k (thR, e)
			  in (thR, mkTyApp (e, x))
		      in doArgs (modTh, ts, tau') k' (thR', e)
*)
(* working version:
			doArgs (modTh, ts, tau') (thR', (thL, mkTyApp (e, ex))) k
*)
(* wrong order:
		      let k' (thR, e) = k (thR, mkTyApp (e, x))
		      in doArgs (modTh, ts, tau') k' (thR', e)
*)
		  | _ -> (* raise (Error ("EVar function type??")) *)
		      error () 


	in let rec findHead (t, ts) = 

	  let cont (x, rho) = 
	    if lev = LTerm then log ("exit_findHead", fun () ->
		   "inf "^String.make i ' '^x^" : "^
		   Clos.toString ((thR, gR), rho)^"\n")
	    else ();
	    doArgs (info, false, ts, rho) (thR, (thL, t)) 
	      (fun ((thR', gR'), thL') -> (thR', thL'))

	  in match t with
	      	C(App, [[],r; [],s]) -> findHead (r, s :: ts)
	      | Var j -> 
		  (if j >= 0 then
		     let (x, rho) = cxt_lookupTmVar (gL, j, gR) 
		     in cont (x, rho)
		   else match env_lookup (thL, j) with
		       Decl (x, rho) -> cont (x, rho)
		     | Def (x, rho, _, _) -> cont (x, rho)
			 (* the following case is needed because
			    chk' is also used to check wellkindedness *)
		     | Cell c -> match !c with
			   Name (x, gX, kappa) -> cont (x, kappa)
			 | Value (tau, gX, kappa) -> cont (Clos.toString ((thL, gX), tau), kappa)
		  )
	      | _ -> raise (Fail ("No typing rule applies to "^
				  Clos.toString ((thL, gL), tL)^" : "^
				  Clos.toString ((thR, gR), tR)^"\n"))

	in (* body chk x @ ts *)
	  findHead (tL, []) 

      

(* else infer [DEPRECATED]
    | _ -> let (theta', tau', e) = inf (i, cL, (thR, gR))
      in let (theta'', tau'', e') = newEVars (false, ((theta', gR), tau'), e)

      in let _ = if verb then
	print_string ("inf "^(String.make i ' ')^
		      (Clos.toString cL)^" : "^
		      (Clos.toString (wnfc ((theta'', gR), tau'')))^"\n")
      else ()

      in 
	Uni.unify (i, (theta'', gR), tau'', tR, mkType); 
	if verb then 
	  print_string ("chk "^(String.make i ' ')^
			(Clos.toString (cL))^" : "^
			(Clos.toString (wnfc ((theta'', gR), tR)))^"\n")
	else ();
	(theta'', e') 
*)

(* and chk  (info, cL, cR) = chk' (info, cL, whnf (cR)) *)
(* and chkL (info, cL, cR) = chk' (info, whnf (cL), cR) *)
		
in (* begin chkVerb' *) 
(*  chk' ((true, 0), wnfc (cL0), wnfc (cR0)) *)
(* without norm: *)
  chk' ((true, 0), cL0, wnfc (cR0))

(* and chkVerb (verb, cL0, cR0) = chkVerb' (verb, cL0, whnf (cR0)) *)

and chkKind (kappa)      = chkVerb' (LKind, false, (Clos.empty, kappa), (Clos.empty, mkKind))
and chkType' (c)         = chkVerb' (LType, false, c, (Clos.empty, mkType))


(* Top-level functions *)

let chkType (c) = chkVerb' (LType, verbWfT, wnfc (c), (Clos.empty, mkType))
let chkConstr (c, kappa) = chkVerb' (LType, verbWfT, wnfc (c), (Clos.empty, kappa))

let chk (((thL, gL), tL) as cL, (((thR, gR), tR) as cR)) =  
  let (thR', (thL', e)) = chkVerb' (LTerm, true, cL, cR) 
  in 
    log ("church_term", fun () ->
	   "\nChurch-Term: "^(Church.toString ((thL', gL), (thR', gR), e))^"\n");
    log ("church_term_deBruijn", fun () ->
	   "C-T[de Bruijn]: "^Exp.toString e^"\n");

    (* double check: try to regain type from Church term *)

    let tau = Church.infer (((thL', gL), e), (thR', gR));
    in 
      log ("inferred_type", fun () ->
	     "Inferred type: "^(Clos.toString ((thR', gR), tau))^"\n");
      Eq.eq ((thR', gR), tR, tau, mkType);

      let e' = Church.etanf ((thL', gL), (thR', gR), e)
      in 
	log ("eta_church_term", fun () ->
	       "Eta-red. Church-Term: "^(Church.toString ((thL', gL), (thR', gR), e'))^"\n");

	let tau' = Church.infer (((thL', gL), e'), (thR', gR));
	in 
	  log ("eta_inferred_type", fun () ->
		 "Inferred type: "^(Clos.toString ((thR', gR), tau'))^"\n");
	  Eq.eq ((thR', gR), tR, tau', mkType);

       (thR', (thL', e'))


