(** A simple lexer for Fomega files *)

open Genlex (* for type token *)

exception Fail of string


(** Simple conversions char list -> string. *)

let implode (l) = 
  let s = String.create (List.length (l)) in
  let rec imp (i, l) = 
    match l with
	[] -> s
      | (c::cs) -> String.set s i c; imp (i+1, cs)
  in
    imp (0, l)

let revImplode (l) = 
  let n = List.length (l) in
  let s = String.create (n) in
  let rec imp (i, l) = 
    match l with
	[] -> s
      | (c::cs) -> String.set s (n-i) c; imp (i+1, cs)
  in
    imp (1, l)


	  
(** Constants to parametrize lexer *)

let isSpace  (c) = String.contains " \t\n" c
let isLetter (c) = (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')
let isDigit  (c) = (c >= '0' && c <= '9')
let isIdChar (c) = isLetter (c) || isDigit (c) || c = '_' || c = '\''
let isIdFirst(c) = isLetter (c) || isDigit (c) || c = '_'

let singleCharTokens = "().;:=\\*"
(* let singleCharTokens = ['('; ')'; '.'; ';'; ':'; '='; '\\'; '*'] *)

let keywords = [Keyword.all; Keyword.tm; Keyword.ty]

type lexresult =
    Just of Genlex.token * (unit -> lexresult)
  | Nothing


(** lexer : char Stream.t -> token Stream.t *)

let rec lexer (s) =
  (* warning: side effects on [s] *)

  let rec lexer' (commDepth) =

    let out (t, k) = Just (t,  k) 

    in let eof () = 
      if commDepth > 0 then raise (Fail "comment exceeds end of file")
      else Nothing
(*
    let out (t, k) = t :: k () 

    in let eof () = 
      if commDepth > 0 then raise (Fail "comment exceeds end of file")
      else [] 
*)
    in let cont (k) =
	try
	  k (Stream.next (s))
	with
	    Stream.Failure -> eof ()
    in let contNonOpt (k) =
	try
	  k (Stream.next (s))
	with
	    Stream.Failure -> raise (Fail "unexpected end of file")
      
  (*  lex : char -> token list *)      
    in let rec lex (c) =

	if c = '}' then contNonOpt (lexRBrace)
	else if c = '%' then cont (lexComment)
	else if isSpace(c) || commDepth > 0 then cont (lex)
	else if isIdFirst(c) then cont (lexId [c])
	else if String.contains singleCharTokens c then
	  out (Kwd (String.make 1 c), fun () -> cont (lex))
	else if c = '-' then contNonOpt (lexArrow)
	else raise (Fail ("illegal character '"^(String.make 1 c)^"' in input"))

    and lexId l c =
      if isIdChar (c) then cont (lexId (c::l))
      else let x = revImplode l
      in if List.mem x keywords then out (Kwd x, fun () -> lex (c))
      else out (Ident x, fun () -> lex (c))

    and lexArrow (c) = 
      if c = '>' then out (Kwd "->", fun () -> cont (lex))
      else raise (Fail ("'>' expected after '-' in input, but found '"^(String.make 1 c)^"'"))

    and lexNewLine (c) =
      if c = '\n' then cont (lex) else cont (lexNewLine)
	  
    and lexComment (c) =
	if c = '{' then lexer' (commDepth+1)
	else cont (lexNewLine)

    and lexRBrace (c) =
      if c = '%' then
	if commDepth > 0 then lexer' (commDepth-1)
	else raise (Fail "spurious end of comment '}%'")
      else raise (Fail ("'%' expected after '}', but found '"^(String.make 1 c)^"'"))
    in (* begin lexer' *)
      cont (lex)

  (* initial lexing function *)
  in let k = ref (fun () -> lexer' (0)) 

  in let nextToken (inputPos) =
      match !k () with
	  Nothing -> None
	| Just (t, k') -> k := k'; Some t

  in (* begin lexer *)
    Stream.from (nextToken)

(*
  let l = lexer' (0)
  in Stream.of_list (l)
*)
(* removed 
    try
    with
	Stream.Failure -> raise (Fail "unexpected end of file")
*)

(*
type Token =
    LPAREN         (** '('      *)
  | RPAREN         (** ')'      *)
  | DOT            (** '.'      *)
  | SEMICOLON      (** ';'      *)
  | COLON          (** ':'      *)
  | EQUALS         (** '='      *)
  | BACKSLASH      (** '\'      *)
  | ARROW          (** '->'     *)
  | STAR           (** '*'      *)
  | FORALL         (** 'forall' *)
  | TERM           (** 'term'   *)
  | TYPE           (** 'type'   *)
  | ID of string   (** starting with a letter, consisting of letter numbers, primes and underscores *)
*)
