(* struct Pie *)

(* PIE systems *)
type 'a pieTm =
    Var of int
  | C of 'a * ((string list * 'a pieTm) list);;

type 'a pieSubst =
    Lift of int
  | Cons of 'a pieTm * 'a pieSubst;;

let emptySubst = Lift 0 ;;
let consSubst (t, theta) = Cons (t, theta) ;;

(* Lift all variables in term r, starting with l, by n     *)
(* Intention: Lift free variables                          *)
(*   If gamma |- t: tau                                    *)
(*   Then gamma, x:sigma |- lift (t, 1) : tau              *)
(* does not lift negative variables in normal usage (l>=0) *)

let rec liftTm (r, l, n) =
  let rec liftBinds ts =
    match ts with
	[] -> []
      | (bs, s) :: ss -> (bs, liftTm (s, l + List.length bs, n)) :: liftBinds ss
  in match r with
      Var k -> if k<l then Var k else Var (k+n)
    | C(a, ts) -> C(a, liftBinds ts);;

let lift (t, n) = liftTm (t, 0, n)

let lift1 (t) = liftTm (t, 0, 1)

(* Abslift: rename variable m to l, lift variables greater l by 1       *)  
(* Intention: copy variable m into variable 0, preparing an abstraction *)

let rec absLiftTm (r, m, l) =
  let rec liftBinds ts =
    match ts with
	[] -> []
      | (bs, s) :: ss -> (bs, absLiftTm (s, m + List.length bs, l + List.length bs)) :: liftBinds ss
  in match r with
      Var k -> if k=m then Var l
      else if k >= l then Var (k+1)
      else Var k
    | C(a, ts) -> C(a, liftBinds ts)
	
let absLift (r, m) = absLiftTm (r, m, 0)
		       
exception Occurrence

let rec lower (r0, n0) =
(* Erase all variables i with l <= i <  n in term r        *)
(* Intention: Lift free variables                          *)
(* Strengthening                                           *)
  let rec iter (r, l, n) =
    let rec lowerBinds ts =
      match ts with
	  [] -> []
	| (bs, s) :: ss -> (bs, iter (s, l + List.length (bs), n + List.length (bs))) :: lowerBinds ss
    in
      match r with
	  Var k -> if k<l then Var k 
	  else if k >= n then Var (k-n+l) else raise Occurrence
	| C(a, ts) -> C(a, lowerBinds ts)
  in
      iter (r0, 0, n0)


let rec liftSubst (theta, n) =
  match theta with
      Lift m -> Lift (m+n)
    | Cons (t, theta') -> Cons (liftTm (t, 0, n), liftSubst (theta', n));;

let rec pickSubst (i, theta) =
  match (i, theta) with
      (k, Lift m) -> Var (k+m) 
    | (0, Cons (t, theta')) -> t
    | (n, Cons (t, theta')) -> pickSubst (n-1, theta');;

let rec extSubst (bs, theta, n) =
  match bs with
      [] -> liftSubst (theta, n)
    | (x :: xs) -> Cons (Var n, extSubst (xs, theta, n+1));;

let rec subst (t, theta) =
  let rec substBinds ts =
    match ts with
	[] -> []
      | ((bs, s) :: ss) -> (bs, subst (s, extSubst (bs, theta, 0)))
	  :: substBinds ss
  in match t with
      Var i -> if i<0 then Var i else pickSubst (i, theta)
    | C(a, ts) -> C(a, substBinds ts);;
      

let subst1 (t, s) =
  subst (t, consSubst (s, emptySubst))

(* rename Variable n to l, shifting down the others *)
(* l is a negative value *)
let rec renVar (l, n, t) =
  let rec renBinds ts =
    match ts with
	[] -> []
      | (bs, s) :: ss -> (bs, renVar (l, n + List.length bs, s)) :: renBinds ss
  in match t with
      Var i -> if i = n then Var l
      else if i > n then Var (i-1)
      else Var i
    | C (a, ts) -> C (a, renBinds ts) 



module IntSet = Set.Make 
		  (struct
		     type t = int
		     let compare = compare
		   end)

type intset = IntSet.t

(** continuation passing style implementation of occurs check *)

let occ (is, t, k, 
	 (d : unit -> 'c)) =   (* continuation if everything is done *)
             
  let rec occ' 
    ((l : int),          (* de Bruijn level                    *)
     (t : 'a pieTm))     (* expression to analyze              *)
    (k : intset -> 'c)   (* continuation if vars are left      *) 
    (is: intset) =       (* set of indices which should occur  *)
    
    match t with
	Var j ->
	  let i = j-l in 
	  let is' = IntSet.remove i is in
	    if IntSet.is_empty is' then d () else k is'
      | C(a, ts) -> 
	  (List.fold_right 
	     (fun (bs, s) -> occ' (l + List.length bs, s)) 
	     ts k) is
	  
  in (* begin occ *)
    if IntSet.is_empty is then d () 
    else occ' (0, t) k is


(** raw printing of de Bruijn terms
 val toString : ('a -> string) -> 'a pieTm -> string *)
let toString const t =
  let rec tts t =
    match t with
	Var i -> string_of_int i
      | C (a, ts) -> const a ^ argsToString ts
  and argsToString ts = 
    let rec ats sep ts =
      match ts with
	  [] -> ""
	| (bs, t) :: ts -> sep ^ bindsToString bs ^ tts t ^ ats ", " ts
    and bindsToString bs =
      match bs with
	  [] -> ""
	| (b :: bs) -> b ^ "." ^ bindsToString bs
    in match ts with
	[] -> ""
      | _ -> ats "(" ts ^ ")"
  in tts t

(* end *)
