(* sig PIE *)

(** Core Expression syntax

  Module [Pie] provides the basic layer terms and types: PIE expressions.
  The name is derived from Felix Joachimski's PIE systems 
  (Permutation-Introduction-Elimination).  We simply use his grammar for
  non-wellfounded terms, but of course we need only the well-founded part.
 *)

(** Each PIE expression is either a variable, denoted by its de Bruijn
index, or a constant with name from type ['a] followed by a list of
arguments, each of which may be preceeded by a list of variables which
are bound in the argument.  Provided variable names ([string]) are
just stems for the actual names in printing.
*)
type 'a pieTm =
    Var of int
  | C of 'a * ((string list * 'a pieTm) list)

exception Occurrence 


type 'a pieSubst
  
val emptySubst : 'a pieSubst
val consSubst  : 'a pieTm * 'a pieSubst -> 'a pieSubst

(** [liftTm (tau, l, n)]: start lifting at [l], lift by [n] *)
val liftTm : 'a pieTm * int * int -> 'a pieTm


 (** [lift (tau, n)] : lift all free variables by n *)
val lift : 'a pieTm * int -> 'a pieTm
val lift1: 'a pieTm -> 'a pieTm

val absLift : 'a pieTm * int -> 'a pieTm

 (** [lower (tau, n)]: strengthen tau by n *)
val lower : 'a pieTm * int -> 'a pieTm
 (* raises exception [Occurrence] if strengthening not possible *)      

val subst  : 'a pieTm * 'a pieSubst -> 'a pieTm
val subst1 : 'a pieTm * 'a pieTm -> 'a pieTm
  (** [subst1 (x.t, s) = [s/x]t] *)

val renVar : int * int * 'a pieTm -> 'a pieTm
      (** [renVar (new, old, t)] *)


module IntSet :
  sig
    type elt = int
    and t
    val empty : t
    val is_empty : t -> bool
    val mem : elt -> t -> bool
    val add : elt -> t -> t
    val singleton : elt -> t
    val remove : elt -> t -> t
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val subset : t -> t -> bool
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val for_all : (elt -> bool) -> t -> bool
    val exists : (elt -> bool) -> t -> bool
    val filter : (elt -> bool) -> t -> t
    val partition : (elt -> bool) -> t -> t * t
    val cardinal : t -> int
    val elements : t -> elt list
    val min_elt : t -> elt
    val max_elt : t -> elt
    val choose : t -> elt
  end

type intset = IntSet.t

(*
module IntSet = Set.Make 
		  (struct
		     type t = int
		     let compare = compare
		   end)

type intset = IntSet.t
*)

(** continuation passing style implementation of occurs check *)
val occ : intset * 'a pieTm * (intset -> 'c) * (unit -> 'c) -> 'c



(** raw printing of deBruijn terms *)
val toString : ('a -> string) -> 'a pieTm -> string
