(* struct strict *)

(** Strictness check for type and constructor definitions

This module implements the strictness check of Pfenning and Schuermann
(TYPES 1998).  A strict notational definition is injective and hence
can be treated like a constant in unification.  Non-strict definitions
have to be expanded every time to ensure soundness.

*)

(*

val strict : exp -> bool

*)

open Output
open Pie
open Exp
open Surr


(** [whnf true t] expands all definitions in [t].
    [whnf false t] expands only non-strict definitions.  
In both cases, [t] may not contain existential variables *)

let whnf expand t0 =
    
    let rec nfApp (s, t) =
      match s with
      	  C (Lam, [[x],r]) -> iter (Pie.subst1 (r, t))
      	| _ -> Exp.mkApp (s, t)
	    
    and iter t =
      match t with
	  Var i -> if i<0 then
	    match env_lookup (Global.tyEnv(), i) with
	        Def (x, kappa, r, strictDef) -> 
		  if expand || not strictDef then iter r else t
	      | _ -> t
	  else t
      	| C (App, [[],r; [],s]) -> nfApp (iter r, s)
      	| _ -> t
    in
      iter t0


let whnf_ = whnf false
let whnfDef = whnf true

let inc (i, j) = (i+1, j+1)
let isDefPar (i, j) k = k < i && k >= j
let defParNum (i, j) k = k - j

(** [remove (a: 'a) (l: 'a list)] removes element [a] from list [l] (order preserving. *)
let remove a l =
  let rec rem l =
    match l with
	[] -> []
      | (x::xs) -> if x=a then xs else x :: rem xs
  in rem l


(** {[pattern (i, ts)]} decides whether the spine [ts] is a pattern,
i.e. each of the arguments [t] in [ts] must be a unique variable with
de Bruijn index [< i] *)
let pattern (i, ts) =
  let rec pat (ts, gone) =
    match ts with
	[] -> true
      | (t :: ts') ->
	  match t with
	      Var k -> if (k >= 0) && (k < i) && not (List.mem k gone)
	      then pat (ts', k::gone)
	      else false
	    | _ -> false
  in pat (ts, [])

(** [strictExp (i, t, xs, k)] tries to find rigid occurrences of some
of the variables [xs] in weak head normal [t] and then continues [k]
with the remaining variables.  If [xs] is empty, [strictExp] succeeds.
Note that to obtain the true de Brujin index of a variable from [xs],
the offset [i] must be added.  Variables below [i] are local
parameters.  Initially, [strictExp] must be called with continuation
[function _ => false] *)
let rec strictExp (i, t, xs, k) =
(*  print_string ("strictExp "^Exp.toString t^"\n");
  flush stdout; *)
  match t with
      C(Lam, [[y],r]) -> strictExp (inc i, whnf_ r, xs, k)
    | C(Forall, [[],kappa; [a],sigma]) -> strictExp (inc i, whnf_ sigma, xs, k)
    | C(Arr, [[],sigma; [],tau]) -> strictVec(i, [sigma; tau], xs, k)
    | C(App, [[],r; [],s]) -> strictApp (i, r, [s], xs, k)
    | Var _ -> strictApp (i, t, [], xs, k)

(** [strictVec (i, ts, xs, k)] calls strictExp for the first
expression in [ts] and shifts the remaining expressions into the
continuation.  The [ts] are not in weak head normal form. *)
and strictVec (i, ts, xs, k) = 
  match xs with
      [] -> true
    | _  -> match ts with 
	  [] -> k xs
	| (t :: ts') -> strictExp (i, whnf_ t, xs, 
				   function xs' -> strictVec (i, ts', xs', k))

(** [strictApp (i, t, ts, xs, k)] works its way to the head of term
[t], which needs to be in weak head normal form. *)
and strictApp ((ii,jj) as i, t, ts, xs, k) =
(*  print_string ("strictApp "^Exp.toString t^"\n");
  flush stdout; *)
  match t with
      C(App, [[],r; [],s]) -> strictApp (i, r, s :: ts, xs, k)
    | Var j -> if isDefPar i j then 
	(let x = defParNum i j
        in if (List.mem x xs) && pattern (jj, ts) then k (remove x xs)
	   else k xs)
      else strictVec (i, ts, xs, k)

  (* no other cases possible for weak head normal terms *)

let rec strictDef (i, t, xs) =
  match t with
      C(Lam, [[x],r]) -> strictDef (i+1, whnfDef r, i::xs)
    | _ -> strictExp ((i, 0), t, xs, function xs' -> false)
(** The failure continuation to [strictExp] will cause [strictApp] to
fail if one of the [xs], let us say [x], matches the head of [t], even
if [x] is the only variable. If [x] does not match, a new continuation
will be created by [strictVec] which will cause the strictness check
to succeed if no more variables [xs] are left. *)

let strict (t) =
  log ("enter_strict", fun () -> "Testing strictness: "); 
  strictDef (0, whnfDef t, [])
