(* sig strict *)

open Exp 

(** Strictness check for type and constructor definitions.

This module implements the strictness check of Pfenning and Schuermann
(TYPES 1998).  A strict notational definition is injective and hence
can be treated like a constant in unification.  Non-strict definitions
have to be expanded every time to ensure soundness.

*)

(** Top level function, decides strictness. *)
val strict : exp -> bool

val whnf : bool -> exp -> exp
(** [whnf true t] expands all definitions in [t].
    [whnf false t] expands only non-strict definitions.  
In both cases, [t] may not contain existential variables *)

val pattern : int * exp list -> bool
(** {[pattern (i, ts)]} decides whether the spine [ts] is a pattern,
i.e. each of the arguments [t] in [ts] must be a unique variable with
de Bruijn index [< i] *)


val strictExp : (int * int) * exp * int list * (int list -> bool) -> bool
(** [strictExp (i, t, xs, k)] tries to find rigid occurrences of some
of the variables [xs] in weak head normal [t] and then continues [k]
with the remaining variables.  If [xs] is empty, [strictExp] succeeds.
Note that to obtain the true de Brujin index of a variable from [xs],
the offset [i] must be added.  Variables below [i] are local
parameters.  Initially, [strictExp] must be called with continuation
[function _ => false] *)


val strictVec : (int * int) * exp list * int list * (int list -> bool) -> bool
(** [strictVec (i, ts, xs, k)] calls strictExp for the first
expression in [ts] and shifts the remaining expressions into the
continuation.  The [ts] are not in weak head normal form. *)


val strictApp : (int * int)* exp * exp list * int list * (int list -> bool) -> bool
(** [strictApp (i, t, ts, xs, k)] works its way to the head of term
[t], which needs to be in weak head normal form. *)

