(* struct surr *)

open Output
open Pie
open Exp

(** COMMON TYPES *)

(** Contexts *)

type preExp = int * exp  (* (cxt level, exp) *)
    
type rawCxt = 
    Empty
  | Cons of string * preExp * rawCxt
      
(** Context : (#TmVars, #TyVars, G) *)
type cxt = int * rawCxt 


(** Environments *)

type cell =
    Name of string * cxt * exp             (** (X, Gamma, kappa) *)
  | Value of exp * cxt * exp               (** (tau, Gamma, kappa) *)
  | ChName of string * cxt * cxt * exp     (** Church cell (x, gL, gR, tau) *)
  | ChValue of exp * cxt * cxt * exp       (** (e, gL, gR, tau) *)

type rawEnv = 
    EEmpty
  | EDecl of string * exp * rawEnv             (* c : tau               *)
  | EDef of string * exp * exp * bool * rawEnv (* c : tau = e (strict?) *)
  | ECell of cell ref * rawEnv
          
type env = int * rawEnv
type entry =
    Decl of string * exp                       (* c : tau     *)
  | Def of string * exp * exp * bool           (* c : tau = e (strict?) *)
  | Cell of cell ref


let env_index ((l, theta), y) =
  let rec iter (i, th) =
    match th with
	EEmpty -> None
      | ECell (c, th') -> 
	  (match !c with
	       Name (x, gamma, kappa) -> if x=y then Some i 
	       else iter (i+1, th')
	     | Value (tau, gamma, kappa) -> iter (i+1, th')
	     | ChName (x, gL, gR, tau) -> if x=y then Some i 
	       else iter (i+1, th')
	     | ChValue (e, gL, gR, tau) -> iter (i+1, th'))
      | EDecl (x, tau, th') -> if x=y then Some i else iter (i+1, th')
      | EDef (x, tau, t, str, th') -> if x=y then Some i else iter (i+1, th')
  in
    iter (l, theta)
      
    
(** CONTEXTS *)

(** Dumping contents of context *)
	     
let cxt_varsToString (n, gamma) =
  let rec iter (i, g) =
    match g with
	Empty -> ""
      | Cons (x, t, g') -> string_of_int i^":"^x^" "^iter (i+1, g')
  in
    iter (0, gamma)
let cxt_dump (g) = "Contents of context: "^cxt_varsToString (g)^"\n"

(** Constructing contexts *)

let cxt_empty = (0, Empty)
let cxt_pushLit (x, t, (n, g)) = (n+1, Cons (x, t, g))

(** push and care for unique names *)

let findInst ((n, gamma), x) =
  let rec iter (g) =
    match g with
	Empty -> None
      | Cons (x', t, g') -> if Names.isInst (x, x') then Some x' else iter (g')
  in
    iter (gamma)

let nextInst ((theta, gamma), x) =
  match findInst (gamma, x) with
      None -> 
	(match env_index (theta, x) with
	     None -> x
	   | Some i -> Names.nextInst (x))
    | Some x' -> Names.nextInst (x')

let cxt_push (x, t, (th, g)) = cxt_pushLit (nextInst ((th, g), x), t, g)
let cxt_pushTmVar (x, ((k, gR), tau), surr) = cxt_push (x, (k, tau), surr)  
let cxt_pushTyVar (a, kappa, surr) = cxt_push (a, (0, kappa), surr)
			    

(** push incomplete entries *)

let dummy = (0, C(Arr, []))      (* some ill-formed pre-exp *)
let cxt_push' (x, surr) = cxt_push (x, dummy, surr)
let cxt_pushLit' (x, g) = cxt_pushLit (x, dummy, g)
		     
let cxt_length (n, g) = n

let pack ((n, g), tau) = (n, tau)
let unpack ((k, tau), (n, g)) =
  if n > k then Pie.lift (tau, n-k)
  else if n < k then raise (Error "Surr.unpack: new cxt is not extension of old ctx")
  else tau
  (* old cxt level : k
     new cxt level : n    Invariant : n >= k
   *)

 (* promote (gamma, t, gamma')
    gamma  old context
    gamma' new context
    raises Pie.Occurrence if strengthening not possible
 *)
let promote ((n, g), t, (n', g')) =
  log ("enter_promote", fun () -> 
	 "promoting "^Exp.toString(t)^" by "^string_of_int (n'-n)^"\n");
  let diff = n'-n 
  in if diff > 0 then Pie.lift (t, diff)
    else if diff < 0 then Pie.lower (t, -diff)
    else t

(* unused
(** Promotion for Church terms *)

let tmVarPromote ((n, g), t, (n', g')) =
(* DEBUG message
  print_string ("promoting "^Exp.toString(t)^" by "^string_of_int (n'-n)^"\n");
*)
  let diff = n'-n 
  in if diff > 0 then Exp.tmVarLift (t, 0, diff)
    else if diff < 0 then raise (Error "strengthening for Church-terms not yet implemented")
    else t

let tyVarPromote ((n, g), t, (n', g')) =
(* DEBUG message
  print_string ("promoting "^Exp.toString(t)^" by "^string_of_int (n'-n)^"\n");
*)
  let diff = n'-n 
  in if diff > 0 then Exp.tyVarLift (t, 0, diff)
    else if diff < 0 then raise (Error "strengthening for Church-terms not yet implemented")
    else t
*)

(* unused
(** Church-style abstraction *)

let tmAbstract ((n, Cons (x, ptau, gL')), gR, e) =
  if n > 0 then ((n-1, gL'), mkTmAbs (x, unpack (ptau, gR), e))
  else raise Error "tmAbstract: cannot abstract in empty context"
*)

(** Lookup functionality *)
		      
let cxt_lookup ((n, gamma), i) =
  let rec iter (m, g) = 
    match g with
	Empty -> 
	  raise (Error ("Surr.cxt_lookup: illformed context.\n"^
			"Actual length does not match claimed length "^
			string_of_int (n)^".\n"^cxt_dump (n, gamma)))
      | Cons (x, t, g') -> if i=m then (x, t) else iter (m+1, g')
  in
    if i >= 0 && i < n then iter (0, gamma)
    else raise (Error ("Surr.cxt_lookup: variable with index "^
		       string_of_int (i)^" not bound in context of (claimed) length "^
		       string_of_int (n)^".\n"^cxt_dump (n, gamma))) 

let cxt_lookupTmVar (gL, i, gR) =
(*
  let _ = print_string ("lookup Var "^string_of_int i^" ==> "); flush stdout in
*)
  let (x, ptau) = cxt_lookup (gL, i) in
  let result = (x, unpack (ptau, gR)) in
(*
  let _ = print_string (x^"\n"); flush stdout in
*)
    result

let cxt_lookupTyVar (gamma, i) =
  let (x, (n, tau)) = cxt_lookup (gamma, i)
  in  (x, tau)

let cxt_index ((n, gamma), y) =
  let rec iter (i, g) =
    match g with
	Empty -> None
      | Cons (x, t, g') -> if x=y then Some i else iter (i+1, g')
  in
    iter (0, gamma)
      



(** ENVIRONMENTS *)

(** Dumping contents of environment *)

let env_varsToString (l, theta) =
  let rec iter (i, th) =
    let cont (th', x) = string_of_int i^":"^x^" "^iter (i+1, th') in
    match th with
	EEmpty -> ""
      | ECell (c, th') -> 
	  (match !c with
	       Name (x, gamma, kappa) -> cont (th', x)
	     | Value (tau, gamma, kappa) -> cont (th', "<value>")
	     | ChName (x, gL, gR, tau) -> cont (th', x)
	     | ChValue (e, gL, gR, tau) ->  cont (th', "<Church-value>"))
      | EDecl (x, tau, th') -> cont (th', x)
      | EDef (x, tau, t, str, th') -> cont (th', x)
  in
    iter (l, theta)
let env_dump (th) = "Contents of environment: "^env_varsToString (th)^"\n"
     
let eVarName (l, a) = "?"^a^string_of_int (-l)
let tmVarName (l, x) = "?"^x^string_of_int (-l)
		   
let env_empty = (0, EEmpty)
	      
let env_pushDecl (x, tau, (l, theta)) = (l-1, EDecl (x, tau, theta))
let env_pushDef  (x, tau, t, strict, (l, theta)) 
  = (l-1, EDef (x, tau, t, strict, theta))
let env_pushEVar (a, gamma, kappa, (l, theta)) = 
  (l-1, ECell (ref (Name (eVarName (l-1, a), gamma, kappa)), theta))
let env_introEVarName (a, gamma, kappa, (l, theta), t) =
  let x = eVarName (l-1, a) in
  ((l-1, ECell (ref (Name (x, gamma, kappa)), theta)),
   Pie.renVar (l-1, 0, t), 
   Var (l-1), 
   x)
let env_introEVarX (a, gamma, kappa, (l, theta), t) =
  (env_pushEVar (a, gamma, kappa, (l, theta)), Pie.renVar (l-1, 0, t), 
   Var (l-1))
let env_introEVar (a, gamma, kappa, (l, theta), t, e) =
  (env_pushEVar (a, gamma, kappa, (l, theta)), Pie.renVar (l-1, 0, t), 
   mkTyApp (e, Var (l-1)))

let env_newChVar (x, gL, gR, sigma, (l, theta)) = 
  let r = ref (ChName (tmVarName (l-1, x), gL, gR, sigma)) in
  ((l-1, ECell (r, theta)), Var (l-1), r)

let env_newTmVar' (x, gamma, sigma, (l, theta)) = 
  let r = ref (Name (tmVarName (l-1, x), gamma, sigma)) in
  ((l-1, ECell (r, theta)), Var (l-1), r)

let env_newEVar' (a, gamma, kappa, (l, theta)) = 
  let r = ref (Name (eVarName (l-1, a), gamma, kappa)) in
  ((l-1, ECell (r, theta)), Var (l-1), r)
 
(* no longer needed: *)
let env_introEVar' (a, gamma, kappa, (l, theta), t, f) =
  (env_pushEVar (a, gamma, kappa, (l, theta)), Pie.renVar (l-1, 0, t), 
   fun es -> mkTyApp (f es, Var (l-1)))
let env_newEVar (a, gamma, kappa, (l, theta)) = 
  ((l-1, ECell (ref (Name (eVarName (l-1, a), gamma, kappa)), theta)), 
   Var (l-1))
 

let env_lookup ((l, theta), i) =
  let rec iter (m, th) =
    match th with
	EEmpty -> 
	  raise (Error ("Surr.env_lookup: illformed environment.\n"^
			"Actual length does not match claimed length "^
			string_of_int (-l)^".\n"^env_dump (l, theta)))
      | ECell (c, th') -> if i=m then Cell c else iter (m+1, th')
      | EDecl (x, tau, th') -> if i=m then Decl (x, tau) else iter (m+1, th')
      | EDef (x, tau, t, strict, th') -> if i=m then Def (x, tau, t, strict) 
	else iter (m+1, th')
  in
    if i < l then 
      raise (Error ("Surr.env_lookup: variable with index "^
		    string_of_int (-i)^" not bound in environment of (claimed) length "^
		    string_of_int (-l)^".\n"^env_dump (l, theta))) 
	
    else iter (l, theta) 
      
let rec env_promote ((l0, th), (lStop, thStop), gNew) =
  let rec iter (l, th) =
    if l = lStop then th
    else match th with
        EEmpty -> th
      | EDecl (x, tau, th') -> EDecl (x, tau, iter (l+1, th'))
      | EDef (x, tau, t, strict, th') -> EDef (x, tau, t, strict, iter (l+1, th'))
      | ECell (c, th') -> 
	  (match !c with
	    Name (x, gamma, kappa) -> c := Name (x, gNew, kappa)
	     | Value (tau, gamma, kappa) -> ());
	  (* do nothing, since tau will be promoted when extracted from th
	     c:= Value (tau, gamma, kappa)); 
	  *)
	  ECell (c, iter (l+1, th'))
  in (l0, iter (l0, th))



(** Normalization procedures. *)

(* These are in here just to use in churchPromote.

   Better solution: implement procedure which just "flattens", i.e.
   resolves instantiated EVars. 
*)

let nf ((expand0, sideterms0), ((theta, gamma), t0)) =
  let modified = ref (false) in 
    
  (* nfDn 
     handle eliminations, unpack Evars *)

  let rec nfDn ((expand, side), g, t0) =
    
    let rec nfApp (s, t) =
      match s with
      	  C (Lam, [[x],r]) -> modified := true; 
	    nfUp ((expand, true), g, Pie.subst1 (r, t))
      	| _ -> if side then Exp.mkApp (s, nfUp ((false, side), g, t))
	  else Exp.mkApp (s, t)
	    
    and iter t =
      match t with
	  Var i -> if i<0 then
	    match env_lookup (theta, i) with
	      	Decl _ -> t
	      | Def (x, kappa, r, strict) -> if expand then iter r else t
	      | Cell c -> match !c with
		    Name _ -> t
		  | Value (r, g', kappa) -> 
		      modified := true; 
		      iter (promote (g', r, g))
      	  else t
      	| C (App, [[],r; [],s]) -> nfApp (nfDn ((expand, side), g, r), s)
      	| _ -> t
    in
      iter t0
      	
  (* nfUp
     handle constructors *)

  and nfUp ((expand, side), g, t0) =
    let t = nfDn ((expand, side), g, t0) in
      if side then
      	match t with
	    C (Lam, [[x],s]) -> mkLam (x, nfUp ((false, side), cxt_push' (x, (theta, g)), s))
      	  | C (Forall, [[],kappa; [a],tau]) -> mkForall (a, kappa, nfUp ((false, side), cxt_pushTyVar (a, kappa, (theta, g)), tau))
      	  | C (Arr, [[],r; [],s]) -> mkArr (nfUp ((false, side),  g, r), nfUp ((false, side), g, s))
      	  | C (Type, []) -> t
      	  | C (Kind, []) -> t
      	  | _ -> t
      else t
  in
(*
  let p1= "nf  "^Clos.toString ((theta, gamma), t0) in
  (* let _ = print_string p1; flush stdout in *)
*)
  let v = nfUp ((expand0, sideterms0), gamma, t0) in
(*
  let p2 = " ==> "^Clos.toString ((theta, gamma), v)^"\n" in
  (* let _ = print_string p2; flush stdout in *)
  let _ = if !modified then print_string (p1^p2) else () in
*)
    v


(* Weak normal form                                                 *)
(* Remove all open redices in a term, but do not expand definitions *)

(* wnfc : clos -> clos *)

(* expand = false, sideterms = true *)
let wnf  (((th, g), t) as c) =           nf ((false, true), c)
let wnfc (((th, g), t) as c) = ((th, g), nf ((false, true), c))


(* Head normal form                                                 *)
(* Expose outermost constructor, expanding definitions as needed    *)

(* Property: if c is in weak normal form, then also hnf c           *)

(* hnfc : clos -> clos *)

(* expand = true, sideterms = false *)
let hnf  (((th, g), t) as c) =           nf ((true, false), c)
let hnfc (((th, g), t) as c) = ((th, g), nf ((true, false), c))




(* unused: *)
let rec curryPromote (((th, ((n, gRaw) as g)), t), ((n', g'Raw) as g')) =

  let diff = n'-n in

  let rec lift (t, l) = 
    match t with
	
	(* variable *)
	Var k -> 
	  (if k >= 0 then 
	     if k < l then Var k else 
	       let k' = k + diff in
		 if k' < 0 then raise (Error "curryPromote: occurrence")
		 else Var k'
	   else match env_lookup (th, k) with
	       Decl _ -> t
	     | Def _ -> t
	     | Cell c -> match !c with
		   Name (x, gX, kappa) -> 
		     raise (Error "curryPromote: do not know how to lift EVars")
		 | Value (tau, gX, kappa) -> 
		     lift (curryPromote (((th, gX), tau), g), l))
	  
        (* others *)
      | C(c, ts) -> C(c, List.map 
			(fun (bs, s) -> (bs, lift (s, l + List.length bs)))
			ts)
  in (* begin curryPromote *)
    lift (t, 0)
(*
    if diff >= 0 then lift (t, 0)  (* hack: if diff=0, expand inst EVars *)
    else raise (Error "curryPromote: strengthening not yet implemented")
*)
	
(* extra effect: weakly normalizes types *)	  
let rec churchPromote (t0, (thL, ((nL, gLraw) as gL)), ((nL', gL'raw) as gL'), 
		           (thR, ((nR, gRraw) as gR)), ((nR', gR'raw) as gR')) =

  let diffL = nL' - nL
  and diffR = nR' - nR in

  log ("enter_churchPromote", fun () ->
	 "churchPromote "^Exp.toString (t0)^", tms by "^string_of_int (diffL)^", tys by "^string_of_int (diffR)^"\n");
 
  let rec lift (t, lL, lR) =
    match t with 

      (* term variable *)
      Var k -> 
	(if k >= 0 then 
	   if k < lL then Var k else 
	     let k' = k + diffL in
	       if k' < 0 then raise (Error "churchPromote: occurrence")
	       else Var k'
	 else match env_lookup (thL, k) with
	     Decl _ -> t
	   | Def _ -> t
	   | Cell c -> match !c with
(*
		 Name (x, gX, tau) -> t
	       | Value (r, gX, tau) -> 
		   let r' = churchPromote (r, (thL, gX), (nL+lL, Empty),
					      (thR, (nR+lR, Empty)), (nR+lR, Empty)) (* bug *) 
*)
		 ChName (x, gLX, gRX, tau) -> t
	       | ChValue (r, gLX, gRX, tau) -> 
		   let r' = churchPromote (r, (thL, gLX), (nL+lL, Empty),
					      (thR, gRX), (nR+lR, Empty)) (* bug *)
		   in lift (r', lL, lR))
(*
raise (Error "lifting for Church terms with term-EVars not yet implemented"))
*)

      (* term var abstraction (count for terms, promote type ann.) *)
    | C(TmAbs, [[],tau; [x],t]) ->
	mkTmAbs (x, 
		 (* remove instantiated EVars, then Pie.liftTm *)
		 liftTm (wnf ((thR, (nR+lR, Empty)), tau), lR, diffR), 
		 lift (t, lL+1, lR))

      (* term application *)
    | C(TmApp, [[],r; [],s]) ->
	mkTmApp (lift (r, lL, lR), lift (s, lL, lR))

      (* type var abstraction (count for types) *)
    | C(TyAbs, [[],kappa; [a],t]) ->
	mkTyAbs (a, kappa, lift (t, lL, lR+1))

      (* type application: right component is tm var free *)
    | C(TyApp, [[],r; [],tau]) ->
	mkTyApp  (lift (r, lL, lR), liftTm (wnf ((thR, (nR+lR, Empty)), tau), lR, diffR))

    | _ -> raise (Error ("ill-formed Church-term"^toString(t0)))

  in (* begin chLift *)
    lift (t0, 0, 0)
(*
    if diffL >= 0 && diffR >= 0 then lift (t0, 0, 0)
    else raise (Error "churchPromote: strengthening not implemented")
*)
