(** Unification *)

open Output
open Pie
open Exp
open Surr

(** Options *)

let indent = Output.indent

(** Exceptions *)

exception Fail of string  (** unification failed *)

(** Canonical forms for unification *)
type uniCf =
    Flex of exp * cell ref    (** (kappa, c)        *)
  | RDef of exp * int * exp   (** (kappa, uid, def) *)
  | Rig of exp                (** (kappa)           *)
  | All of string * exp * exp (** (a, kappa, tau)   *)
  | Arrow of exp * exp        (** (sigma, tau)      *)
  | Abs

let kind (f : uniCf) =
  match f with
      Flex (kappa, _) -> Some kappa
    | RDef (kappa, _, _) -> Some kappa
    | Rig (kappa) -> Some kappa
    | All _ -> Some mkType
    | Arrow _ -> Some mkType
    | Abs -> None

type spine = uniCf * exp * exp list (** [(f, t, l)] *) 
type spineCl = Clos.surr * spine

let tailToString (surr, l) = 
  List.fold_left (fun str -> fun t -> str ^ Clos.toString' (10, (surr, t)) ^ " ") "" l

let spineToString (surr, (f, t, l)) =
  match f with
      Flex (kappa, c) -> 
	(match (!c) with
	     Name (x, g, kappa) -> "Flex "^x^" "^tailToString (surr, l)
	   | _ -> raise (Error ("spineToString: hit Flex with already instantiated cell")))
    | RDef (kappa, i, d) -> "Def "^string_of_int i^"[="^Clos.toString (surr, d)^"] "^tailToString (surr, l)
    | Rig (kappa) -> "Rig "^Clos.toString (surr, t)^" "^tailToString (surr, l)
    | Arrow (s1, s2) -> "Arr "^Clos.toString' (5, (surr, s1))^" -> "^Clos.toString' (4, (surr, s2))
    | All (a, k, t) -> "All "^a^":"^Exp.toString k^". "^Clos.toString (Clos.pushTyVar (a, k, surr) , t)
    | Abs -> "Abs "^Clos.toString (surr, t)

let liftSpine (f, t, l) =
  let t' = Pie.lift1 t
  and l' = List.map Pie.lift1 l
  in match f with
      Flex _ -> (f, t', l')
    | RDef (kappa, i, d) -> (RDef (kappa, i, Pie.lift1 d), t', l')
    | Rig _ -> (f, t', l')
    | All (a, kappa, tau) -> (All (a, kappa, Pie.liftTm (tau, 1, 1)), t', l')
    | Arrow (r, s) -> (Arrow (Pie.lift1 r, Pie.lift1 s), t', l')
    | Abs -> (f, t', l')

(* This is an instance of List.fold_left 
let rec toExp' (t, l) =
  match l with
      [] -> t
    | (x :: xs) -> toExp' (mkApp (t, x), xs)
*)
let toExp' (t, l) = List.fold_left (fun r -> fun s -> mkApp (r,s)) t l
let toExp (f, t, l) = toExp' (t, l)

(** Test canonical spines, whether both 
let spineEqFlex ((t1, l1), (t2, l2)) =
  try 
    Eq.eq (t1, t2)
*)

(** General purpose function splitAt *)

let rec splitAt (n, l) =
  if n<=0 then ([], l)
  else match l with
      [] -> ([], [])
    | x :: xs -> let (h, t) = splitAt (n-1, xs) in (x :: h, t)
  

(** Compute a canonical form (head normal) for unification.

[canf : clos * exp list -> spine]

[canf (((th, g), head), tail) = (form, kappa, head', tail')]

Input parameters:
 - [(th, g) : surr ] ** full environment 
 - [head : exp     ] ** head of input spine
 - [tail : exp list] ** tail of input spine

Output parameters:
 - [form  : uniCf  ] ** [Flex], [RDef], [Rig], [All] or [Arrow]
 - [kappa : exp    ] ** kind of [head']
 - [head' : exp    ] ** new head
 - [tail' : exp lis] ** new tail

Invariants:
 - [head] @ [tail] -->beta [head'] @ [tail']
 - [head'] is neither an abstraction (in case of input of kind "*") nor an application nor an instantiated EVar
 - [head'] @ [tail'] is no redex unless [head'] is a [RDef]
 - if [form = All(a,kappa,tau)] then [t = forall a:kappa.tau] and [l = nil]
 - if [form = Arrow(sigma,tau)] then [t = sigma -> tau] and [l = nil]
 - if [form = Abs] then [l = nil]
 - if [form = Flex] then [t = Var k] for some [k] < 0
*)

let canf (((th, g), t0), l0) =

  let error () = raise (Error ("canf: not a type")) in

  let rec iter (t, l) =
    let return (f) = (f, t, l) 
    in
      match t with
	  Var i -> 
	    if i >= 0 then 
	      let (a, kappa) = cxt_lookupTyVar (g, i)
	      in return (Rig (kappa))
	    else (match env_lookup (th, i) with
		      Decl (x, kappa) -> return (Rig (kappa))
		    | Def (x, kappa, tau, strict) -> 
			(* Maybe it is wiser to never expand definitions, 
			   be they strict or non-strict
			   But then one could not use definitions which ignore one of their arguments.
			   Currently "strictness" is reduced to simple occ.

			  return (RDef (kappa, i, tau))
			*)
			if strict then 
			  return (RDef (kappa, i, tau))
			else iter (tau, l)
		    | Cell c -> match !c with
		      	  Name (x, gX, kappa) -> return (Flex (kappa, c))
		      	| Value (tau, gX, kappa) -> iter (promote (gX, tau, g), l))
      	| C(App, [[],r; [],s]) -> iter (r, s :: l)
      	| C(Lam, [[x],r]) ->  (match l with
				   [] -> return (Abs)
				 | (s :: l') -> iter (Pie.subst1 (r, s), l'))
	| C(Forall, [[],kappa; [a],tau]) -> 
	    if l = [] then return (All (a, kappa, tau))
	    else error ()
	| C(Arr, [[],sigma; [],tau]) ->
	    if l = [] then return (Arrow (sigma, tau))
	    else error ()
      	| _ -> return (Rig (mkType))
  in
    iter (t0, l0)


(** [isFlexPlusArgs ((th, g), tau)] returns [true] 
  iff [tau] = ?X sigma_0 ... sigma_n *)
let isFlexPlusArgs (cl) =
  let (f, t, l) = canf (cl, []) in
    match f with
	Flex _ -> List.length (l) > 0
      | _ -> false


(* Abstraction *)

(* assume l0 to be a list of variables; abstract all *)
let abstract (c1, ((thR, gR), t0), l0) =

  let fail () = 
    match !c1 with
	Name (x, _, _) ->
	  raise (Fail ("Cannot abstract variables to solve the equation\n"^
		       (List.fold_left 
			  (fun x -> fun r -> x^" "^Clos.toString' (10, ((thR,gR),r)))
			  x l0)^
		       " = "^Clos.toString ((thR, gR), t0)^".\n"^
                       "The arguments to "^x^" need to be pairwise distinct local variables.\n"))
      | _ -> raise (Error ("abstract.error: unification variable is not uninstantiated."))
  in 

  let rec iter (n, t, l) =
    match l with
      	[] -> t
      | (r :: rs) -> 
	  let fail () = (* if argument is not variable, insert void binding *)
	    mkLam ("_", iter (n+1, Pie.lift1 (t), rs))
	  in
	  match r with
	    Var i -> 
	      if i >= 0 then 
		(* make var i+n into var 0, lift all others *)
		let t' = Pie.absLift (t, i + n) 
		and (x, _) = cxt_lookup (gR, i)
		in mkLam (x, iter (n+1, t', rs))
	      else 
		(match env_lookup (thR, i) with
		     Def _ -> fail () 
		   | Decl _ -> fail ()
		   | Cell c -> match !c with
			 Value (tau, gX, kappa) -> 
			   iter (n, t, promote (gX, tau, gR) :: rs)
		       | _ -> fail ())
	  | _ -> fail ()

  in (* begin abstract *)
    iter (0, t0, l0)
      
(* Assign cell *)		

let assign (ind, c, ((theta, gamma), rho)) =
  match !c with
      Name (x, gX, kappa) ->
	log ("set_uni", fun () ->
	     "set "^(String.make ind ' ')^x^" = "^
	     Clos.toString ((theta, gamma), rho)^"\n");
	let rho' = 
	  try
	    promote (gamma, rho, gX)
	  with Pie.Occurrence -> raise (Fail ("Cannot perform assignment. Right hand side contains local variables which may not occurr in EVar "^x^".\n"))
	in
	  c := Value (rho', gX, kappa)
    | _ -> raise (Error ("assign: trying to reassign unification variable"))


(** Unification procedure *)

(** Types for CPS-style unification and type checking 
type state = env * (env * exp)
type cont  = state -> state
*)

type uol =
    Unify of exp  (* kind [kappa] *)
  | Subty of Clos.clos 
      (* Church-term [((thL, gL), e)] *)

(** [uniOrLeq : uol * int * env * (env * cxt) * exp * exp  -> cxt * (env * exp)]
    [thStop]: where to stop post-lifting.  This probably has to change in the Arrow-rule
 *)
let uniOrLeq (uol, ind, thStop, (theta, gamma), sigma, tau) : cxt * (env * exp) =

(* THIS ALGORITHM IS QUITE STUPID (2002-08-27)
** E.g. 
**
**     |- X26   = \A. mu2g J Phi (H A) : * -> *
** u:* |- X26 u = mu2g J Phi (H u)     : *
** u:* |- X26   = mu2g J Phi           : *       FAILS!
** 
** Is one side is just a flexible variable, then the solution is clear anyway!
**
** PROBLEM SOLVED (2003-08-25)
*)

  let spinesDontMatch ((thR, gR), s1, s2) =
    raise (Fail ("Clash: head symbols do not match.\n"^
		 "LHS: "^(Clos.toString ((thR, gR), toExp (s1)))^"\n"^
                 "RHS: "^(Clos.toString ((thR, gR), toExp (s2)))^"\n"))
  in

(* Kind-driven unification *)

  (* to do: uniUp should use spines 
     problem: we only implemented spines for types (kind * ),
     hence, the head of a spine was never an abstraction
     this would change if we used spines in uniUp
   *)
  let rec uniUp (th, g, r1, r2, k) =
    log ("enter_uniUp", fun () ->
	   "uup "^Clos.toString ((th, g), r1)^" == "^Clos.toString ((th, g), r2)^" : "^Clos.toString (Clos.close (k))^"\n");

    let rec uniUp' (g, ((f1, t1, l1) as s1), ((f2, t2, l2) as s2), k) =
      (* case: one side flexible *)
      match (f1, f2) with
	  (Flex (k1, c), _) -> flex (ind, th, g, (c, k1, t1, l1), s2)
	| (_, Flex (k2, c)) -> flex (ind, th, g, (c, k2, t2, l2), s1)
	    
	(* else: break down to kind "*" *)
	| _ -> match k with
	      C(Arr, [[],k1; [],k2]) -> 
		  let g' = cxt_pushTyVar ("u", k1, (th, g)) 
		  and (f1', t1', l1') = liftSpine (s1) 
		  and (f2', t2', l2') = liftSpine (s2) in
		    uniUp' (g',
			    canf (((th, g'), t1'), List.append l1' [Var 0]), 
			    canf (((th, g'), t2'), List.append l2' [Var 0]), 
		            k2)
	    | _  -> uniDn (th, g, s1, s2)

    in (* begin uniUp *)
      uniUp' (g, canf (((th, g), r1), []), canf (((th, g), r2), []), k)
(*
    match k with
	C(Arr, [[],k1; [],k2]) -> 
	  let g' = cxt_pushTyVar ("u", k1, (th, g))
	  in uniUp (th, g', mkApp (Pie.lift1 (s), Var 0), 
		            mkApp (Pie.lift1 (t), Var 0), k2)
      | _  -> uniDn (th, g, canf (((th, g), s), []), canf (((th, g), t), []))
*)

  and uniVec (th, g, l1, l2, k) =
    match (k, l1, l2) with
	(_, [], []) -> ()
      |	(C(Arr, [[],k1; [],k2]), x::xs, y::ys) ->
	  uniUp (th, g, x, y, k1);
	  uniVec (th, g, xs, ys, k2)
      | _ -> raise (Error ("uniVec: kind does not match lists")) 

(* Heuristic for instantiating EVars *)

  and flex (ind, th, g, (c1, k1, t1, l1), ((f2, t2, l2) as s2)) =

    log ("enter_flex", fun () ->
	   "flx "^spineToString ((th, g), (Flex (k1, c1), t1, l1))^
           " == "^spineToString ((th, g), s2)^"\n");

    let n1 = List.length (l1) 
    and n2 = List.length (l2) in

    let flexL () =

      let pointwise () =
	let 
	  (l2a, l2b) = splitAt (n2-n1, l2)
	in 
	  assign (ind, c1, ((th, g), toExp' (t2, l2a)));
       	  uniVec (th, g, l1, l2b, k1)

      and abstr () =
	assign (ind, c1, ((th, g), abstract (c1, ((th, g), toExp (s2)), l1)))

      in (* begin flexL *)
	if n1 = n2 then
	  match kind (f2) with
	      Some k2 ->
		(* test whether heads have same kind *)
		(try
		   Eq.streq (Clos.empty, k1, k2);
		   (* yes, then unify pointwise *)
		   pointwise ()
		 with Eq.Fail _ ->
		   (* no, abstract *)
		   abstr ())
	    | None -> pointwise ()
	else if n1 < n2 then 
	  pointwise ()
	else abstr ()

    in (* begin flex *)
    (* test wether s2 is Flex with same flexible variable *)
    match f2 with
	Flex (k2, c2) ->
	  (* test whether same flexible variable *)
	  (try
	     Eq.streq ((th, g), t1, t2);
	     let r1 = toExp' (t1, l1)
	     and r2 = toExp' (t2, l2) in
	       (* test whether s1 = s2 *)
	       (try
		  (* if yes, nothing to do *)
		  Eq.streq ((th, g), r1, r2)
		with Eq.Fail _ -> 
		  (* if no, fail *)
		  raise (Fail ("do not know how to solve equation "^
			       Clos.toString ((th, g), r1)^" = "^
			       Clos.toString ((th, g), r2))))
	   with Eq.Fail _ ->
	     (* not the same flexible variable: continue.
                Bug: There might be still problems with circular assignments
		of flexible variables. *)
	     
	     (* assign flexible variable with fewer arguments *)
	     if n1 <= n2 then 
	       flexL ()
	     else 
	       let (l1a, l1b) = splitAt (n1-n2, l1)
	       in
		 assign (ind, c2, ((th, g), toExp' (t1, l1a)));
		 uniVec (th, g, l2, l1b, k2))

      | _ -> (* second spine not flexible *)
	  flexL ()

(* Structural unification *)

  and uniDn (th, g, ((f1, t1, l1) as s1), ((f2, t2, l2) as s2)) =
    match (f1, f2) with
	(Flex (k1, c), _) -> flex (ind, th, g, (c, k1, t1, l1), s2)
      | (_, Flex (k2, c)) -> flex (ind, th, g, (c, k2, t2, l2), s1)
      | (RDef (k1, i1, d1), RDef (k2, i2, d2)) -> 
	  if i1 = i2 then uniVec (th, g, l1, l2, k1)
	  else uniDn (th, g, canf (((th, g), d1), l1), s2)
      | (RDef (k1, i1, d1), _) -> uniDn (th, g, canf (((th, g), d1), l1), s2) 
      | (_, RDef (k2, i2, d2)) -> uniDn (th, g, s1, canf (((th, g), d2), l2)) 
      | (All (a1, kappa1, tau1), All (a2, kappa2, tau2)) ->
	  uniRig (th, g, kappa1, kappa2);
	  let g' = cxt_pushTyVar (a1, kappa1, (th, g))
	  in uniUp (th, g', tau1, tau2, mkType)	  
      | (Arrow (sig1, sig2), Arrow (tau1, tau2)) ->
	  uniUp (th, g, sig1, tau1, mkType);
	  uniUp (th, g, sig2, tau2, mkType) 
      | (Rig (k1), Rig (k2)) ->
	  uniRig (th, g, t1, t2); 
	  uniVec (th, g, l1, l2, k1)
      | _ -> spinesDontMatch ((th, g), s1, s2)

  and uniRig (th, g, s, t) =
    match (s, t) with
	(Var i, Var j) -> if i=j then () 
	else raise (Fail ("Clash: inequal bound variables "^Clos.toString ((th, g), s)^" and "^Clos.toString ((th, g), t)^".\n"))
      | (C(Arr, [[],s1;[],s2]), C(Arr, [[],t1;[],t2])) ->
	  uniUp (th, g, s1, t1, mkType);
	  uniUp (th, g, s2, t2, mkType);
(*
      | (_, C(Forall, [[], k2; [a2],t2])) ->
	  let g' = cxt_pushTyVar (a2, k2, (theta, g))
	  in uniUp (th, g', promote (g, s, g'), t2, mkType)
      | (C(Forall, [[],k1; [a1],t1]), _) ->
          let e = mkType in  (* TODO: Church terms *)
	  let (th', s', e') = env_introEVar (a1, g, k1, th, t1, e)
	  in uniUp (th', g, s', t, mkType)
*)
(* Functionality moved to uniDn
      | (C(Forall, [[],k1; [a1],t1]), C(Forall, [[], k2; [a2],t2])) ->
	  uniRig (th, g, k1, k2);
	  let g' = cxt_pushTyVar (a1, k1, (th, g))
	  in uniUp (th, g', t1, t2, mkType)
*)
      | (C(Type, []), C(Type, [])) -> ()
      | (C(Kind, []), C(Kind, [])) -> ()
      | _ -> raise (Fail ("Clash: "^Clos.toString ((th, g), s)^" does not match "^Clos.toString ((th, g), t)^".\n"))
  in


  (** Subtyping *)

  (** This is a top level function---there is no call back to [leq] from
     the other unification procedures.

     [leq: env -> (bool * int) * cxt * (env * cxt) * exp -> spine * spine -> cxt * env * exp]

     [leq thL (info, gL, (thR, gR), e) (s1, s2) = (gR', thR', e')]

     Input: 
     [((thL, gL), (thR, gR), e) : Church.clos]
     [((thR, gR), s1)] : left spine 
     [((thR, gR), s2)] : right spine 
     Output:
     [gR'] extended type context to check remaining args in e (exec k)
     [((thL, gL), (thR', gR), e') : Church.clos] Note that only thR' is ext.
     Invariants:
     [thL] remains untouched
    *)

  let leq (thL) =

    let rec leq' ((first, i), gL, (thR, gR), e) =

      let rec leq'' (((f1, t1, l1) as s1), ((f2, t2, l2) as s2)) : cxt * (env * exp) =

	log ("enter_leq", fun () ->
	       "leq "^(String.make i ' ')^spineToString ((thR, gR), s1)^
	       " <= "^spineToString ((thR, gR), s2)^"\n");
	log ("church_uni", fun () ->
	       "      "^(String.make ind ' ')^"[Church-in: "^Church.toString ((thL, gL), (thR, gR), e)^"]\n"); 
	
	let return (gR, (thR, e)) = (gR, (thR, e)) in
	let return' () = return (gR, (thR, e)) in
	  
	  match (f1, f2) with
	      (Flex (k1, c), _) -> flex (i, thR, gR, (c, k1, t1, l1), s2); return' ()
	    | (_, Flex (k2, c)) -> flex (i, thR, gR, (c, k2, t2, l2), s1); return' ()
		
	    (* forallR *)
	    | (_, All (a2, kappa2, tau2)) ->
		
		(* introduce new type variable in context, lift Church-term
		   and EVars *)
		let newI = if first then i+indent else i in 
		let gR' = cxt_pushTyVar (a2, kappa2, (thR, gR)) in
		  log ("ass_uni", fun () ->
			 "ass "^(String.make newI ' ')^
			 (Clos.toString ((thR, gR'), Var 0))^" : "^
			 (Clos.toString (Clos.close (kappa2)))^"\n");
		  let e' = churchPromote (e, (thL, gL), gL, (thR, gR), gR')
		  and s1' = liftSpine s1 in
		    (* the presence of the following stmnt. has no real effect?
		       [at least for all examples in testsuite]:
		    *)
		    ignore (env_promote (thR, thStop, gR')); 
		    (* promote 'til env theta *)
		    
		    let (gR1, (thR1, e1)) = 
		      leq' ((false, newI), gL, (thR, gR'), e')
			(s1', canf (((thR, gR'), tau2), [])) 
		    in return (gR1, (thR1, mkTyAbs (a2, kappa2, e1))) 
			 
            (* forallL *)
	    | (All (a1, kappa1, tau1), _) -> 
		let newI = if first then i+indent else i in
		let (thR', tau1', ex, x) = 
		  env_introEVarName (a1, gR, kappa1, thR, tau1) 
		in 
		  log ("new_uni", fun () -> "new "^(String.make newI ' ')^x^"\n");
		  leq' ((false, newI), gL, (thR', gR), mkTyApp (e, ex))
		    (canf (((thR', gR), tau1'), []), s2)


  	    (* arrow *)
	    | (Arrow (sig1, sig2), Arrow (tau1, tau2)) ->
		(* we introduce a new variable for the Church term,
		   hence later need to lift [e] (see [promote])
		*)
		let newI = if first then i+indent else i in
		let gLx = cxt_pushTmVar ("x", (gR, tau1), (thL, gL)) in
		  log ("ass_uni", fun () -> "ass "^(String.make newI ' ')^
			 (Clos.toString ((thL, gLx), Var 0))^" : "^
			 (Clos.toString ((thR, gR), tau1))^"\n");

		  (* 1st step: tau1 <= sig1 
		     ignore extensions to gR *)
		  let (_, (thR, ex)) =
		    leq' ((true, newI), gLx, (thR, gR), Var 0) 
		      (canf (((thR, gR), tau1), []),
	               canf (((thR, gR), sig1), [])) in

        	  (* 2nd step: sig2 <= tau2 
		     pass on extensions to gR *)
		  let (gR', (thR, e')) =
		    leq' ((false, newI), gLx, (thR, gR), 
			  mkTmApp (churchPromote (e, (thL, gL), gLx, 
						     (thR, gR), gR), ex))
		      (canf (((thR, gR), sig2), []),
	               canf (((thR, gR), tau2), [])) in

		    (* 3rd step: put abstraction onto Church term *)
		    return (gR', (thR, mkTmAbs ("x", tau1, e')))
	          (* or: Church.tmAbs'' ((thL, gLx), (thR, gR), e') *)

            (* Heuristics for definitions: 
             - treat all definitions as invariant
             - expand def. first which was introduced earlier *)
	    | (RDef (k1, i1, d1), RDef (k2, i2, d2)) ->
		if i1 = i2 then (uniVec (thR, gR, l1, l2, k1); return' ())
		else if i1 < i2 then leq'' (canf (((thR, gR), d1), l1), s2)
		else leq'' (s1, canf (((thR, gR), d2), l2))

	    | (_, RDef (k2, i2, d2)) -> 
		log ("debug_leq", fun () ->
			"leq: expanding rhs def "^spineToString ((thR, gR), s2)^"\n"); 
		leq'' (s1, canf (((thR, gR), d2), l2))
	    | (RDef (k1, i1, d1), _) -> 	  
		log ("debug_leq", fun () ->
		       "leq: expanding lhs def "^spineToString ((thR, gR), s1)^"\n");
		leq'' (canf (((thR, gR), d1), l1), s2)

	    | (Rig (k1), Rig (k2)) -> 
		uniRig (thR, gR, t1, t2);
		uniVec (thR, gR, l1, l2, k1); 
		return' ()

	    | _ -> spinesDontMatch ((thR, gR), s1, s2)

      in (* begin leq' *)
	leq''
    in (* begin leq *)
      leq'

  in (* begin [uniOrLeq] *)
    match uol with
	Unify (kappa) -> uniUp (theta, gamma, sigma, tau, kappa);
	  (* make up bogus ret. value *) (gamma, (theta, Var 0))
      | Subty ((thL, gL), e) -> leq thL ((false, ind), gL, (theta, gamma), e) 
	  (canf (((theta, gamma), sigma), []), 
	   canf (((theta, gamma), tau), [])) 


(* begin comment CPS-style unification

  in let rec leq ((first, i) as info, gL, g, 
		  ((f1, t1, l1) as s1), 
		  ((f2, t2, l2) as s2)) 
      ((th, ((thL, e) as ecl)):state) (k:cont) =
      
      log ("enter_leq", fun () ->
	     "leq "^(String.make i ' ')^spineToString ((th, g), s1)^
	     " <= "^spineToString ((th, g), s2)^"\n");
       log ("church_uni", fun () ->
	      "      "^(String.make ind ' ')^"[Church-in: "^Church.toString ((thL, gL), (th, g), e)^"]\n"); 
       

      (* generally, the continuation k is executed in the end in all cases *)

    match (f1, f2) with
	(Flex (k1, c), _) -> flex (i, th, g, (c, k1, t1, l1), s2); k (th, ecl)
      | (_, Flex (k2, c)) -> flex (i, th, g, (c, k2, t2, l2), s1); k (th, ecl)

	  (* forallR *)
      | (_, All (a2, kappa2, tau2)) ->

          (* introduce new type variable in context, lift Church-term
             and EVars *)
	  let newI = if first then i+indent else i in 
	  let g' = cxt_pushTyVar (a2, kappa2, (th, g)) in
	    log ("ass_uni", fun () ->
		   "ass "^(String.make newI ' ')^
		   (Clos.toString ((th, g'), Var 0))^" : "^
		   (Clos.toString (Clos.close (kappa2)))^"\n");
	  let e' = churchPromote (e, (thL, gL), gL, (th, g), g')
	  and s1' = liftSpine s1
	  in 
	    (* the presence of the following statement has no real effect 
	       [at least for all examples in testsuite]:
*)
	       ignore (env_promote (th, theta, g')); 
	       (* promote 'til env theta *)

	    (* 1st step: s1' <= tau2 *)
	     leq ((false, newI), gL, g', s1', canf (((th, g'), tau2), [])) 
	       (th, (thL, e'))
	       (fun (th', (thL', e')) ->

		  (* 2nd step: add tyAbs onto Church term *)
		  (* 3rd step: k *)
		  k (th', (thL', mkTyAbs (a2, kappa2, e'))))
(* TODO: pass extended context g' to k *)

(* k before mkTyAbs -- does NOT work
	       (fun (th', e'cl) ->

		  (* 2nd step: k *)
		  let (th', (thL', e')) =  k (th', e'cl) in

		    (* 3rd step: add tyAbs onto Church term *)
		    (th', (thL', mkTyAbs (a2, kappa2, e'))))
*)

(* forallR -- Original version
      | (_, All (a2, kappa2, tau2)) ->
	  let g' = cxt_pushTyVar (a2, kappa2, (th, g))
	  and s1' = liftSpine s1
	  in ignore (env_promote (th, theta, g')); (* promote 'til env theta *)
	     let (th', e') = leq (g', s1', canf (((th, g'), tau2), [])) (th, e)
	     in (th', mkTyAbs (a2, kappa2, e'))
*)

	  (* forallL -- CPS style *)	  
      | (All (a1, kappa1, tau1), _) ->
	  let newI = if first then i+indent else i in
	  let (th', tau1', ex, x) = env_introEVarName (a1, g, kappa1, th, tau1)
	  in
	    log ("new_uni", fun () -> "new "^(String.make newI ' ')^x^"\n");
	    leq ((false, newI), gL, g, canf (((th', g), tau1'), []), s2) 
	       (th', (thL, mkTyApp (e, ex))) k

(* forallL -- Original version	  
      | (All (a1, kappa1, tau1), _) ->
	  let (th', tau1', e') = env_introEVar (a1, g, kappa1, th, tau1, e)
	  in leq (g, canf (((th', g), tau1'), []), s2) (th', e')
*)

	  (* arrow -- CPS style *)
      | (Arrow (sig1, sig2), Arrow (tau1, tau2)) ->
	  (* we introduce a new variable for the Church term,
             hence later need to lift [e] (see [promote])
           *)
	  let newI = if first then i+indent else i in
          let gLx = cxt_pushTmVar ("x", (g, tau1), (thL, gL)) in
	    log ("ass_uni", fun () -> "ass "^(String.make newI ' ')^
		   (Clos.toString ((thL, gLx), Var 0))^" : "^
		   (Clos.toString ((th, g), tau1))^"\n");
	  (* 1st step: tau1 <= sig1 *)
	  leq ((true, newI), gLx, g, canf (((th, g), tau1), []),
	               canf (((th, g), sig1), [])) (th, (thL, Var 0))
	  (fun (th1, (thL1, ex)) -> 

	     (* 2nd step: sig2 <= tau2 *)
	     leq ((false, newI), gLx, g, canf (((th1, g), sig2), []),
	                  canf (((th1, g), tau2), [])) 
	     (th1, (thL1, mkTmApp (churchPromote (e, (thL, gL), gLx, (th, g), g), ex))) 
             (* or: (th1, (thL1, mkTmApp (Pie.lift1 e, ex))) *)
	     (* here we assume that thL1 is an extension of thL *)

(* first tmAbs then k [works]
*)
	     (fun (th2, (thL2, e')) -> 

		(* 3rd step: put abstraction onto Church term *)
                (* 4th step: k *)
		k (th2, (thL2, mkTmAbs ("x", tau1, e')))))
	          (* or: Church.tmAbs'' ((thL2, gLx), (th2, g), e')
		     here we assume that gL2 = gL, x:tau1 *)


(* first k then tmAbs [does not work]
	     (fun (th2, e'cl) -> 

		(* 3rd step: k *)
		let (th2, (thL2, e')) = k (th2, e'cl) in

		  (* 4th step: put abstraction onto Church term *)
		  (th2, (thL2, mkTmAbs ("x", tau1, e')))))
	          (* or: Church.tmAbs'' ((thL2, gLx), (th2, g), e')
		     here we assume that gL2 = gL, x:tau1 *)
*)

(* arrow -- Original version
      | (Arrow (sig1, sig2), Arrow (tau1, tau2)) ->
	  (* we have no context for the Church-Term available,
             so we just forget about the name "x", instead just lift [e] 
             let g' = cxt_pushTmVar ("x", (g, tau1), (th, g))
           *)
	  let (th1, ex) = leq (g, canf (((th, g), tau1), []),
			          canf (((th, g), sig1), [])) (th, Var 0)
	  in let (th2, e') = leq (g,
				  canf (((th1, g), sig2), []),
				  canf (((th1, g), tau2), [])) 
			       (th1, mkTmApp (Pie.lift1 e, ex))
	  in (th2, mkTmAbs ("x", tau1, e'))
*)

          (* Heuristics for definitions: 
             - treat all definitions as invariant
             - expand rhs def. first (no specific reason) *)
      | (RDef (k1, i1, d1), RDef (k2, i2, d2)) -> 
	  if i1 = i2 then (uniVec (th, g, l1, l2, k1); k (th, ecl))
	  else if i1 < i2 then leq (info, gL, g, canf (((th, g), d1), l1), s2) (th, ecl) k
	  else leq (info, gL, g, s1, canf (((th, g), d2), l2)) (th, ecl) k
      | (_, RDef (k2, i2, d2)) -> 
	  log  ("debug_leq", fun () -> 
		  "leq: expanding rhs def "^spineToString ((th, g), s2)^"\n"); 
	  leq (info, gL, g, s1, canf (((th, g), d2), l2)) (th, ecl) k
      | (RDef (k1, i1, d1), _) -> 	  
	  log ("debug_leq", fun () ->
		 "leq: expanding lhs def "^spineToString ((th, g), s1)^"\n");
	  leq (info, gL, g, canf (((th, g), d1), l1), s2) (th, ecl) k

(* Definitions -- original version
           (* Heuristics for definitions: 
             - treat all definitions as invariant
             - expand rhs def. first (no specific reason) *)
      | (RDef (i1, d1), RDef (i2, d2)) -> 
	  if i1 = i2 then (uniVec (th, g, l1, l2, k1); (th, e))
	  else if i1 < i2 then leq (g, canf (((th, g), d1), l1), s2) (th, e)
	  else leq (g, s1, canf (((th, g), d2), l2)) (th, e)
      | (_, RDef (i2, d2)) -> 
(*	  print_string ("leq: expanding rhs def "^spineToString ((th, g), s2)^"\n"); *)
	  leq (g, s1, canf (((th, g), d2), l2)) (th, e)
      | (RDef (i1, d1), _) -> 	  
(*	  print_string ("leq: expanding lhs def "^spineToString ((th, g), s1)^"\n"); *)
	  leq (g, canf (((th, g), d1), l1), s2) (th, e) 
*)

     | (Rig (k1), Rig (k2)) -> 
	  uniRig (th, g, t1, t2);
	  uniVec (th, g, l1, l2, k1); k (th, ecl)
      | _ -> raise (SpinesDontMatch (((th, g), s1), ((th, g), s2)))

  in (* body of [uniOrLeq] *)
    match uol with
	Unify (kappa) -> uniUp (theta, gamma, sigma, tau, kappa);
	  (* make up bogus ret. value *) (theta, (env_empty, Var 0))
      | Subty (((thL, gL), e), k) -> leq ((false, ind), gL, gamma, 
			  canf (((theta, gamma), sigma), []), 
			  canf (((theta, gamma), tau), [])) (theta, (thL, e)) k

end comment CPS-style unification *)

(** [leq (info, gL, gamma, sigma, tau) (theta, e) k = (theta', e')] 

Input parameters:
 - [ind   : int]  ** Indentation for instantiation messages 
 - [gamma : cxt]  ** Context for [sigma] and [tau]
 - [sigma : exp]  ** Subtype
 - [tau   : exp]  ** Supertype
 - [k     :cont]  ** Continuation (what to do after unification?)
 - [theta : env]  ** Initial environment, maybe extended
 - [ecl  : clos]  ** Initial Church term (as Clos, clos), maybe extended

Output parameters:
 - [theta': env]  ** Possibly extended theta
                     and, of course, with new instantiations
 - [e'cl : clos]  ** Possibly eta-expanded form of e'      

Description: 

[leq] tries to show that [sigma] is a subtype of [tau] by
instantiating unification variables in [theta].  If unification fails,
an exception is raised.  The initial Church-term [e] is extended by
type applications and abstractions, if necessary.

The purpose of [leq] is to handle universal quantification
appropriately.  While 

  [forall X sigma == forall X tau  <==>  sigma == tau]

holds for logical equivalence [==] and hence is implemented in
unification,  it is too strong a requirement for subtyping [<=].  Here
we have the [forallR] and [forallL] rules:

  forall X.  sigma <= tau
  ----------------------- forallR
  sigma <= (forall X tau)

  exists X.  sigma <= tau
  ----------------------- forallL .
  (forall X sigma) <= tau

Furthermore, we have subtyping for arrow:

   t1 <= s1    s2 <= t2
   -------------------- arrow .
   s1 -> s2 <= t1 -> t2

With Church-terms, the rules read as follows:

  X:kappa |- e : sigma  <=  e': tau
  ------------------------------------------------ forallR
  e : sigma  <=  \\X:kappa.e' : forall X:kappa tau

  e X : sigma  <=  e' : tau
  ---------------------------------------- forallL
  e : (forall X:kappa sigma)  <=  e' : tau

  x:t1 |- x : t1 <= ex : s1    x:t1 |- e ex : s2 <= e' : t2
  --------------------------------------------------------- arrow .
  e : s1 -> s2  <=  \x:t1.e' : t1 -> t2

Implementation:  

Rule [forallL] makes us introduce new EVars into [theta] which may be
instantiated by terms valid in context [gamma].  Rule [forallR] makes
us introduce an new parameter into [gamma].  Thus, we need to post-lift
new EVars to new contexts.  Hence, we need to mark where
we started introducing new EVars into [theta].  We implement this by
keeping a second record of new EVars besides the general environment
[theta]. 
*)    

(* unify : int * (env * cxt) * exp * exp * exp -> unit *)
let unify (ind, (theta, gamma), sigma, tau, kappa) =
  ignore (uniOrLeq (Unify (kappa), ind, theta, (theta, gamma), sigma, tau))

let subType (ind, thStop) ((thL, gL), (thR, gR), e) (sigma, tau) =
  uniOrLeq (Subty ((thL, gL), e), ind, thStop, (thR, gR), sigma, tau)
